/* 
   $Id: main.c,v 1.4 2001/04/23 04:09:27 sho Exp $

   Copyright (C) 2001 AXE,Inc. <sg@axe-inc.co.jp>

   This file is part of sgwm.

   sgwm is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   sgwm is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with sgwm; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "sgwm.h"
#include <X11/X.h>
#include <gtk/gtk.h>
#include <stdlib.h>
#include <stdio.h>
#include <glib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <libintl.h>
#include "atoms.h"
#include "sgwmwidget.h"
#include "sgimbar.h"

#define RCFILENAME ".sgwmrc"

static char *us_rcfile = NULL;

static void usage(char *cmd)
{
    g_error("Usage:%s [-rcfile rcfilename][-middle keycode][-right keycode][-nunome]\n",cmd);
    exit(2);
}

static void parse_args(int argc, char **argv)
{
    int i;

    for (i = 1;i < argc;i++) {
	if (argv[i][0] == '-') {
	    if (strcmp(argv[i]+1,"middle") == 0) {
		if (++i >= argc) {
		    usage(argv[0]);
		}
		middleButtonKeycode = atoi(argv[i]);
	    } else if (strcmp(argv[i]+1,"right") == 0) {
		if (++i >= argc) {
		    usage(argv[0]);
		}
		rightButtonKeycode = atoi(argv[i]);
	    } else if (strcmp(argv[i]+1,"nunome") == 0) {
		nunome = !nunome;
	    } else if (strcmp(argv[i]+1,"rcfile") == 0) {
		if (++i >= argc) {
		    usage(argv[0]);
		}
		us_rcfile = argv[i];
	    } else {
		usage(argv[0]);
	    }
	} else {
	    usage(argv[0]);
	}
    }
}

static gboolean canread(char *fn)
{
    int fd;

    if ((fd = open(fn, O_RDONLY)) < 0) return FALSE;
    close(fd);
    return TRUE;
}

static void exec_rcfile()
{
    char *p;
    char *path = 0;

    if (us_rcfile != NULL) {
	if (!canread(us_rcfile)) {
	    g_error("Can't read rcfile:%s\n",us_rcfile);
	    exit(2);
	}
	path = strdup(us_rcfile);
    }
    if (path == NULL && (p = g_get_current_dir()) != NULL) {
	path = g_strdup_printf("%s/%s",p,RCFILENAME);
	if (!canread(path)) {
	    g_free(path);
	    path = 0;
	}
	g_free(p);
    }
    if (path == NULL && (p = g_get_home_dir()) != NULL) {
	path = g_strdup_printf("%s/%s",p,RCFILENAME);
	if (!canread(path)) {
	    g_free(path);
	    path = 0;
	}
    }
    if (path == NULL) {
	path = g_strdup_printf("%s/%s",SYSCONFDIR,RCFILENAME);
	if (!canread(path)) {
	    g_free(path);
	    path = 0;
	}
    }
    if (path != NULL) {
	char *cmd = g_strdup_printf(". %s",path);
	g_free(path);
	system(cmd);
	g_free(cmd);
    }
}

int main(int argc, char **argv)
{
    GtkWidget *wm;

    gtk_set_locale();
    gtk_init(&argc,&argv);
    parse_args(argc,argv);
    bindtextdomain(PACKAGE, LOCALEDIR);
    textdomain(PACKAGE);
    intern_atoms();
    wm = GTK_WIDGET(sgwmwidget_new());
    gtk_widget_show(wm);
    sgwmwidget_start(SGWMWIDGET(wm));
    exec_rcfile();
    gtk_main();
    gtk_exit(0);
    return 0;
}
