/* 
   $Id: sgbarholder.c,v 1.4 2001/07/04 07:33:12 sho Exp $

   Copyright (C) 2001 AXE,Inc. <sg@axe-inc.co.jp>

   This file is part of sgwm.

   sgwm is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   sgwm is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with sgwm; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "sgwm.h"
#include <gdk/gdkx.h>
#include "sgwmwidget.h"
#include "sgbarholder.h"
#include "sgbar.h"
#include "sgwindow.h"
#include "sgdefmenubar.h"
#include "sgdeftoolbar.h"
#include "sgimbar.h"

static void sgbarholder_class_init(SGBarHolderClass *klass);
static void sgbarholder_init(SGBarHolder *win);
static GtkContainerClass *parent_class;

GtkType sgbarholder_get_type(void)
{
    static GtkType sgbarholder_type = 0;

    if (sgbarholder_type == 0) {
	GtkTypeInfo sgbarholder_info = {
	    "SGBarHolder",
	    sizeof(SGBarHolder),
	    sizeof(SGBarHolderClass),
	    (GtkClassInitFunc)sgbarholder_class_init,
	    (GtkObjectInitFunc)sgbarholder_init,
	    NULL,
	    NULL,
	    NULL
	};
	sgbarholder_type = gtk_type_unique(gtk_container_get_type(),
	    &sgbarholder_info);
    }

    return sgbarholder_type;
}

static void sgbarholder_draw(GtkWidget *widget, GdkRectangle *area)
{
    SGBarHolder *barholder;
    GdkRectangle child_area;

    barholder = SGBARHOLDER(widget);
    if (area == NULL) {
	gtk_widget_draw(barholder->menubarContainer,NULL);
	gtk_widget_draw(barholder->toolbarContainer,NULL);
	gtk_widget_draw(barholder->imbarContainer,NULL);
    } else {
	if (gtk_widget_intersect(barholder->menubarContainer,area,
	  &child_area)) {
	    gtk_widget_draw(barholder->menubarContainer,&child_area);
	}
	if (gtk_widget_intersect(barholder->toolbarContainer,area,
	  &child_area)) {
	    gtk_widget_draw(barholder->toolbarContainer,&child_area);
	}
	if (gtk_widget_intersect(barholder->imbarContainer,area,
	  &child_area)) {
	    gtk_widget_draw(barholder->imbarContainer,&child_area);
	}
    }
}

static void sgbarholder_show(GtkWidget *widget)
{
    SGBarHolder *barholder;

    barholder = SGBARHOLDER(widget);
    GTK_WIDGET_CLASS(parent_class)->show(widget);
    gtk_widget_show(barholder->menubarContainer);
    gtk_widget_show(barholder->toolbarContainer);
    gtk_widget_show(barholder->imbarContainer);
}

static void sgbarholder_realize(GtkWidget *widget)
{
    SGBarHolder *barholder;

    GTK_WIDGET_CLASS(parent_class)->realize(widget);
    barholder = SGBARHOLDER(widget);
    gtk_widget_realize(barholder->menubarContainer);
    gtk_widget_realize(barholder->toolbarContainer);
    gtk_widget_realize(barholder->imbarContainer);
}

static void sgbarholder_size_allocate(GtkWidget *widget,
	GtkAllocation *allocation)
{
    GtkAllocation alloc;
    SGBarHolder *barholder;
    GtkRequisition childReq;
    GtkRequisition imbarReq;
    gint x;
    gint imbarx;
    SGBar *bar;

    barholder = SGBARHOLDER(widget);
    widget->allocation = *allocation;

    gtk_widget_size_request(barholder->menubarContainer,&childReq);
    alloc.x = 0;
#ifdef __SIKIGAMI_PDA__
    alloc.y = allocation->y;
#else
    alloc.y = 0;
#endif
    alloc.width = childReq.width;
    alloc.height = allocation->height;
    gtk_widget_size_allocate(barholder->menubarContainer,&alloc);

    gtk_widget_size_request(barholder->toolbarContainer,&childReq);
    gtk_widget_size_request(barholder->imbarContainer,&imbarReq);
    imbarx = widget->allocation.width-imbarReq.width;
    bar = SGBAR(barholder->toolbarContainer);
    if (sgbar_get_x(bar,&x)) {
	alloc.x = x;
	sgbar_set_movable(bar,TRUE);
    } else {
	alloc.x += alloc.width;
	if (childReq.width > 0 && alloc.x+childReq.width > imbarx) {
	    sgbar_set_movable(bar,TRUE);
	    if (alloc.x > imbarx-SG_BAR_HANDLE_WIDTH) {
		alloc.x = imbarx-SG_BAR_HANDLE_WIDTH;
	    }
	} else {
	    sgbar_set_movable(bar,FALSE);
	}
    }
    alloc.width = childReq.width;
    gtk_widget_size_allocate(barholder->toolbarContainer,&alloc);

    alloc.x = imbarx;
    alloc.width = imbarReq.width;
    gtk_widget_size_allocate(barholder->imbarContainer,&alloc);
}

static void sgbarholder_size_request(GtkWidget *widget,
	GtkRequisition *req)
{
    GtkRequisition childReq;
    SGBarHolder *barholder;
    barholder = SGBARHOLDER(widget);

    childReq.width = 0;
    childReq.height = 0;
    req->width = 0;
    req->height = SG_BARHOLDER_DEFAULT_HEIGHT;
    gtk_widget_size_request(barholder->menubarContainer,&childReq);
    req->width += childReq.width;
    req->height = MAX(req->height, childReq.height);
    gtk_widget_size_request(barholder->toolbarContainer,&childReq);
    req->width += childReq.width;
    req->height = MAX(req->height, childReq.height);
    gtk_widget_size_request(barholder->imbarContainer,&childReq);
    req->width += childReq.width;
    req->height = MAX(req->height, childReq.height);
}

static void sgbarholder_set_default(SGBarHolder *barholder)
{
    GtkBin *bin;

    if (barholder->defaultMenubar != NULL) {
	bin = GTK_BIN(barholder->menubarContainer);
	if (bin->child == NULL) {
	    gtk_container_add(GTK_CONTAINER(barholder->menubarContainer),
	      barholder->defaultMenubar);
	    gtk_widget_show(barholder->defaultMenubar);
	}
    }
    if (barholder->defaultToolbar != NULL) {
	bin = GTK_BIN(barholder->toolbarContainer);
	if (bin->child == NULL) {
	    gtk_container_add(GTK_CONTAINER(barholder->toolbarContainer),
	      barholder->defaultToolbar);
	    gtk_widget_show(barholder->defaultToolbar);
	}
    }
}

void sgbarholder_add(GtkContainer *container, GtkWidget *widget)
{
    SGWindow *sgwin;
    GtkWidget *menubar, *toolbar;
    SGBarHolder *barholder;

    barholder = SGBARHOLDER(container);
    if (widget == NULL) {
	sgbarholder_set_default(barholder);
	return;
    }
    sgwin = SGWINDOW(widget);
    switch (sgwin->type) {
    case SGWINDOW_APP:
	sgwindow_get_bars(sgwin,&menubar,&toolbar);
	if (menubar == NULL && toolbar == NULL) {
	    sgbarholder_set_default(barholder);
	} else {
	    sgbarholder_clear(container);
	    if (menubar != NULL) {
		gtk_container_add(GTK_CONTAINER(barholder->menubarContainer),menubar);
		gtk_widget_show(menubar);
	    }
	    if (toolbar != NULL) {
		gtk_container_add(GTK_CONTAINER(barholder->toolbarContainer),toolbar);
		gtk_widget_show(toolbar);
	    }
	}
	break;
    default:
	break;
    }
}

void sgbarholder_clear(GtkContainer *container)
{
    SGBarHolder *barholder = SGBARHOLDER(container);
    GtkBin *bin;

    bin = GTK_BIN(barholder->menubarContainer);
    if (bin->child != NULL) {
	gtk_widget_hide(bin->child);
	gtk_container_remove(GTK_CONTAINER(barholder->menubarContainer),
	  bin->child);
    }
    bin = GTK_BIN(barholder->toolbarContainer);
    if (bin->child != NULL) {
	gtk_widget_hide(bin->child);
	gtk_container_remove(GTK_CONTAINER(barholder->toolbarContainer),
	  bin->child);
    }
}

void sgbarholder_remove(GtkContainer *container, GtkWidget *widget)
{
    sgbarholder_clear(container);
}

static void sgbarholder_class_init(SGBarHolderClass *klass)
{
    GtkWidgetClass *widget_class;
    GtkContainerClass *container_class;

    parent_class = gtk_type_class(gtk_container_get_type());
    widget_class = (GtkWidgetClass *)klass;
    widget_class->show = sgbarholder_show;
    widget_class->draw = sgbarholder_draw;
    widget_class->realize = sgbarholder_realize;
    widget_class->size_allocate = sgbarholder_size_allocate;
    widget_class->size_request = sgbarholder_size_request;
    container_class = (GtkContainerClass *)klass;
    container_class->add = sgbarholder_add;
    container_class->remove = sgbarholder_remove;
}

static void sgbarholder_init(SGBarHolder *holder)
{
    GtkWidget *widget = GTK_WIDGET(holder);

    GTK_WIDGET_SET_FLAGS(widget,GTK_NO_WINDOW);
    holder->menubarContainer = sgbar_new();
    gtk_widget_set_parent(holder->menubarContainer,widget);
    holder->toolbarContainer = sgbar_new();
    gtk_widget_set_parent(holder->toolbarContainer,widget);
    holder->imbarContainer = sgbar_new();
    gtk_widget_set_parent(holder->imbarContainer,widget);
    holder->defaultMenubar = NULL;
    holder->defaultToolbar = NULL;
    holder->imbar = NULL;
}

GtkWidget *sgbarholder_new(SGWMWidget *sgwmwidget)
{
    GtkWidget *widget;
    SGBarHolder *holder;

    widget  = GTK_WIDGET(gtk_type_new(sgbarholder_get_type()));
    holder = (SGBarHolder *)widget;
    holder->sgwm = sgwmwidget;
    if ((holder->defaultMenubar = sgdefmenubar_new(sgwmwidget)) != NULL) {
	gtk_widget_ref(holder->defaultMenubar);
	gtk_container_add(GTK_CONTAINER(holder->menubarContainer),
	  holder->defaultMenubar);
	gtk_widget_show(holder->defaultMenubar);
    }
    if ((holder->defaultToolbar = sgdeftoolbar_new(sgwmwidget)) != NULL) {
	gtk_widget_ref(holder->defaultToolbar);
	gtk_container_add(GTK_CONTAINER(holder->toolbarContainer),
	  holder->defaultToolbar);
	gtk_widget_show(holder->defaultToolbar);
    }
    if ((holder->imbar = sgimbar_new(sgwmwidget)) != NULL) {
	gtk_widget_ref(holder->imbar);
	gtk_container_add(GTK_CONTAINER(holder->imbarContainer),holder->imbar);
	gtk_widget_show(holder->imbar);
    }
    return widget;
}

void sgbarholder_get_x(GtkWidget *widget, gint *x)
{
    SGBarHolder *holder = SGBARHOLDER(widget);
    gint tx;
    
    gtk_widget_get_pointer(widget,&tx,NULL);
    if (tx < 0) {
	*x = 0;
    } else if (tx > holder->imbarContainer->allocation.x-SG_BAR_HANDLE_WIDTH) {
	*x = holder->imbarContainer->allocation.x-SG_BAR_HANDLE_WIDTH;
    } else {
	*x = tx;
    }
}
