/* 
   $Id: sgwmtop.c,v 1.18 2001/07/05 04:06:19 sho Exp $

   Copyright (C) 2001 AXE,Inc. <sg@axe-inc.co.jp>

   This file is part of sgwm.

   sgwm is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   sgwm is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with sgwm; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "sgwm.h"
#include <gdk/gdkx.h>
#include <gtk/gtkhpaned.h>
#ifdef __SIKIGAMI_PDA__
#include <gtk/gtkvpaned.h>
#endif
#include <gtk/gtkframe.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkviewport.h>
#include "sgwmtop.h"
#include "sgwindow.h"
#include "sgworkspace.h"
#include "sgbarholder.h"
#include "atoms.h"


static void sgwmtop_class_init(SGWMTopClass *klass);
static void sgwmtop_init(SGWMTop *sgwm);

GtkType sgwmtop_get_type(void)
{
    static GtkType sgwmtop_type = 0;

    if (sgwmtop_type == 0) {
	GtkTypeInfo sgwmtop_info = {
	    "SGWMTop",
	    sizeof(SGWMTop),
	    sizeof(SGWMTopClass),
	    (GtkClassInitFunc)sgwmtop_class_init,
	    (GtkObjectInitFunc)sgwmtop_init,
	    NULL,
	    NULL,
	    NULL
	};
	sgwmtop_type = gtk_type_unique(gtk_container_get_type(),
	    &sgwmtop_info);
    }

    return sgwmtop_type;
}

static void sgwmtop_show(GtkWidget *widget)
{
    SGWMTop *sgwm;

    GTK_WIDGET_SET_FLAGS(widget,GTK_VISIBLE);
    gtk_widget_map(widget);
    sgwm = SGWMTOP(widget);
    if (sgwm->bar) gtk_widget_show(sgwm->bar);
    if (sgwm->area) gtk_widget_show(sgwm->area);
}

static void sgwm_paned_set_position(GtkPaned *paned, gint position)
{
    if(position < paned->min_position) position = paned->min_position;
    if(position > paned->max_position) position = paned->max_position;
    if(paned->position_set && paned->child1_size == position) return;
    gtk_paned_set_position(paned, position);
}

#define PANED_HDL_SZ	10
static void paned_chk_handle(SGWMTop *top)
{
    gint n, lastOne, sz;
    GtkPaned *paned = GTK_PANED(top->area);

    sz = 0;
    n = sgwmtop_get_enabled_workspace_count(top, &lastOne);
    if(n > 1) sz = PANED_HDL_SZ;

    if(paned->handle_size == sz) return;
    gtk_paned_set_handle_size(paned, sz);
}

static gboolean sgwmtop_expose(GtkWidget *widget, GdkEventExpose *event)
{
    GdkRectangle child_area;
    SGWMTop *top = SGWMTOP(widget);

    if (top->bar != NULL && gtk_widget_intersect(top->bar,
      &event->area,&child_area)) {
	gtk_widget_draw(top->bar,&child_area);
    }
    if (top->area != NULL && gtk_widget_intersect(top->area,
      &event->area,&child_area)) {
	gtk_widget_draw(top->area,&child_area);
    }
    return TRUE;
}

static void workspace_disable(SGWMTop *top, gint ws)
{
    gint np;
    if (ws == 0) {
	np = 0;
    } else {
#ifdef __SIKIGAMI_PDA__
	np = GTK_WIDGET(top)->allocation.height;
#else
	np = GTK_WIDGET(top)->allocation.width;
#endif
    }
    sgwm_paned_set_position(GTK_PANED(top->area),np);
}

static void nunome_set_position(SGWMTop *top, SGWindow *win, gint no, gint add)
{
    gint pos;
    gint x,y,w,h,bw;

    sgwindow_get_geometry(win,&x,&y,&w,&h,&bw);
#ifdef __SIKIGAMI_PDA__
    if (no == 0) {
	h += 8;
	pos = h;
    } else {
	h += 16;
	pos = top->area->allocation.height - h + add;
    }
#else
    w += NUNOME_EXTRA_SIZE;
    if (no == 0) {
	pos = w;
    } else {
	pos = top->area->allocation.width - w + add;
    }
#endif
    sgwm_paned_set_position(GTK_PANED(top->area),pos);
}

static void sgwmtop_check_resize(GtkContainer *container)
{
    SGWMTop *sgwm;
    GtkRequisition req;
    GtkAllocation alloc;
    GtkWidget *widget;

    sgwm = SGWMTOP(container);
    widget = GTK_WIDGET(container);
    alloc.x = alloc.y = 0;
    if (sgwm->bar) {
	gtk_widget_size_request(sgwm->bar,&req);
#ifdef __SIKIGAMI_PDA__
	alloc.y = widget->allocation.height - req.height;
#endif
	alloc.width = widget->allocation.width;
	alloc.height = req.height;
	gtk_widget_size_allocate(sgwm->bar,&alloc);
    }
    if (sgwm->area) {
#ifdef __SIKIGAMI_PDA__
	alloc.y = 0;
#else
	alloc.y = req.height;
#endif
	alloc.height = widget->allocation.height - req.height;
#ifdef __SIKIGAMI_PDA__
	{ 
	    gint i;
	    SGWorkSpace *ws;

	    for(i=0; i<SG_N_WORKSPACE; i++){
		if(sgwm->workspace[i] == NULL) continue;
		ws = SGWORKSPACE(sgwm->workspace[i]);
		if(ws->nunome){
		    nunome_set_position(sgwm, sgwm->curTask[i], i,
		      alloc.height - sgwm->area->allocation.height);
		}
	    }
	}
#endif
	gtk_widget_size_allocate(sgwm->area,&alloc);
    }
    if (sgwm->bar) {
	gtk_widget_draw(sgwm->bar,NULL);
    }
    if (sgwm->area) {
	gtk_widget_draw(sgwm->area,NULL);
    }
#ifdef __SIKIGAMI_PDA__
    { 
	gint i;
	SGWorkSpace *ws;

	for(i=0; i<SG_N_WORKSPACE; i++){
	    if(sgwm->workspace[i] == NULL) continue;
	    ws = SGWORKSPACE(sgwm->workspace[i]);
	    if(!ws->nunome && ws->disabled){
		if(sgwm->area != NULL) workspace_disable(sgwm, i);
	    }
	}
    }
#endif
}

static void sgwmtop_size_request(GtkWidget *widget, GtkRequisition *req	)
{
    widget->requisition.width = widget->allocation.width;
    widget->requisition.height = widget->allocation.height;
    if (req != NULL) {
	*req = widget->requisition;
    }
}

static void sgwmtop_size_allocate(GtkWidget *widget, GtkAllocation *alp)
{
    GtkAllocation alloc;
    GtkRequisition req;
    SGWMTop *sgwm;

    if (widget->allocation.x == alp->x
      && widget->allocation.y == alp->y
      && widget->allocation.width == alp->width
      && widget->allocation.height == alp->height) {
	return;
    }
    widget->allocation = *alp;
    if (GTK_WIDGET_REALIZED(widget)) {
	gdk_window_move_resize(widget->window,alp->x,alp->y,alp->width,alp->height);
	sgwm = SGWMTOP(widget);
	req.width = req.height = 1;
	gtk_widget_size_request(sgwm->bar,&req);
	alloc.x = alloc.y = 0;
#ifdef __SIKIGAMI_PDA__
	alloc.y = widget->allocation.height - req.height;
#endif
	alloc.width = widget->allocation.width;
	alloc.height = req.height;
	gtk_widget_size_allocate(sgwm->bar,&alloc);
#ifdef __SIKIGAMI_PDA__
	alloc.y = 0;
#else
	alloc.y = req.height;
#endif
	alloc.height = widget->allocation.height - req.height;
	gtk_widget_size_allocate(sgwm->area,&alloc);
	gtk_widget_draw(sgwm->bar,NULL);
	gtk_widget_draw(sgwm->area,NULL);
    }
}

static void sgwmtop_do_activate_workspace(SGWMTop *top, gint no,
  gboolean warp, gboolean focus)
{
    if (no < 0) {
	int i;

	for (i = 0;i < SG_N_WORKSPACE;i++) {
	    SGWorkSpace *ws = SGWORKSPACE(top->workspace[i]);
	    if (!ws->disabled && !ws->nunome) {
		no = i;
		break;
	    }
	}
    }
    if (top->activeWorkSpace >= 0 && top->activeWorkSpace != no) {
	sgbarholder_clear(GTK_CONTAINER(top->bar));
	gtk_widget_set_state(top->workspace[top->activeWorkSpace],
	  GTK_STATE_NORMAL);
	sgwindow_set_workspace_property(NULL,no);
    }
    top->activeWorkSpace = no;
    if (no >= 0) {
	if (top->curTask[no] != NULL) {
	    sgbarholder_add(GTK_CONTAINER(top->bar),GTK_WIDGET(top->curTask[no]));
	} else {
	    sgbarholder_add(GTK_CONTAINER(top->bar),NULL);
	}
	gtk_widget_set_state(top->workspace[top->activeWorkSpace],
	  GTK_STATE_SELECTED);
	if (warp) {
	    sgworkspace_warp_pointer(SGWORKSPACE(
	      top->workspace[top->activeWorkSpace]));
	}
	if (focus) sgwindow_set_focus(top->curTask[no]);
    } else {
	sgwindow_set_workspace_property(NULL,0);
	sgwindow_set_focus(NULL);
    }
    gtk_widget_queue_resize(top->bar);
}

void sgwmtop_activate_workspace(SGWMTop *top, gint no, gboolean warp,
  gboolean focus)
{
    SGWorkSpace *ws;

    if (no >= SG_N_WORKSPACE) return;
    if (no >= 0) {
	ws = SGWORKSPACE(top->workspace[no]);
	if (ws->disabled || ws->nunome) return;
    }
    if (no < 0) {
	if (top->activeWorkSpace < 0 && top->curTask[0] != NULL) {
	    no = 0;
	} else {
	    int i;

	    i = top->activeWorkSpace+1;
	    if (i >= SG_N_WORKSPACE) i = 0;
	    while (i != top->activeWorkSpace && (top->curTask[i] == NULL
	      || SGWORKSPACE(top->workspace[i])->nunome)) {
		i++;
		if (i >= SG_N_WORKSPACE) i = 0;
	    }
	    if (i == top->activeWorkSpace) {
		return;
	    } else {
		no = i;
	    }
	}
    }
    if (top->activeWorkSpace != no) {
	sgwmtop_do_activate_workspace(top,no,warp,focus);
    }
}

static void sgwmtop_realize(GtkWidget *widget)
{
    GdkWindowAttr attributes;
    gint attributes_mask;
    SGWMTop *sgwm;
    GtkAllocation alloc;
    GtkContainer *con;
    GtkWidget *frame;
    GtkRequisition req;
    GtkWidget *sw[SG_N_WORKSPACE];
    int i;

    if (GTK_WIDGET_REALIZED(widget)) return;
    GTK_WIDGET_SET_FLAGS(widget,GTK_REALIZED);

    attributes.x = widget->allocation.x;
    attributes.y = widget->allocation.y;
    attributes.width = widget->allocation.width;
    attributes.height = widget->allocation.height;
    attributes.window_type = GDK_WINDOW_TOPLEVEL;
    attributes.wclass = GDK_INPUT_OUTPUT;
    attributes.visual = gtk_widget_get_visual(widget);
    attributes.colormap = gtk_widget_get_colormap(widget);
    attributes.event_mask = GDK_EXPOSURE_MASK;
    attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
    widget->window = gdk_window_new(NULL,&attributes,attributes_mask);
    gdk_window_set_user_data(widget->window, widget);
    widget->style = gtk_style_attach(widget->style, widget->window);
    gtk_style_set_background(widget->style, widget->window, widget->state);

    sgwm = SGWMTOP(widget);
    sgwm->bar = sgbarholder_new(sgwm->wmwidget);
#ifdef __SIKIGAMI_PDA__
    sgwm->area = gtk_vpaned_new();
#else
    sgwm->area = gtk_hpaned_new();
#endif
    gtk_widget_set_parent(sgwm->bar,widget);
    gtk_widget_set_parent(sgwm->area,widget);
    req.width = req.height = 1;
    gtk_widget_size_request(sgwm->bar,&req);
    alloc.x = alloc.y = 0;
#ifdef __SIKIGAMI_PDA__
    alloc.y = widget->allocation.height - req.height;
#endif
    alloc.width = widget->allocation.width;
    alloc.height = req.height;
    gtk_widget_size_allocate(sgwm->bar,&alloc);
    con = GTK_CONTAINER(sgwm->area);

    for (i = 0;i < SG_N_WORKSPACE;i++) {
	sgwm->workspace[i] = sgworkspace_new(sgwm->wmwidget,NULL,NULL,i);
	sw[i] = gtk_scrolled_window_new(NULL,NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw[i]),
	  GTK_POLICY_AUTOMATIC,GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(sw[i]),sgwm->workspace[i]);
	sgwm->frame[i] = frame = gtk_frame_new(NULL);
	gtk_widget_ref(frame);
	gtk_frame_set_shadow_type(GTK_FRAME(frame),GTK_SHADOW_IN);
	gtk_container_add(con,frame);
	gtk_container_add(GTK_CONTAINER(frame),sw[i]);
	gtk_widget_show(frame);
	gtk_widget_show(sgwm->workspace[i]);
	gtk_widget_show(sw[i]);
    }

    gtk_widget_realize(sgwm->bar);
#ifdef __SIKIGAMI_PDA__
    alloc.y = 0;
#else
    alloc.y = req.height;
#endif
    alloc.height = widget->allocation.height - req.height;
    gtk_widget_size_allocate(sgwm->area,&alloc);
#ifdef __SIKIGAMI_PDA__
    sgwm_paned_set_position(GTK_PANED(sgwm->area),SG_MAINAREA_SIZE(alloc.height));
#else
    sgwm_paned_set_position(GTK_PANED(sgwm->area),SG_MAINAREA_SIZE(alloc.width));
#endif

#ifdef __SIKIGAMI_PDA__
    sgwm->frame[0]->allocation.height = 168;
    sgwmtop_enable_workspace(sgwm, 1, False);
#endif
    gtk_widget_realize(sgwm->area);
    sgwmtop_do_activate_workspace(sgwm,0,TRUE,TRUE);
}

static void sgwmtop_class_init(SGWMTopClass *klass)
{
    GtkWidgetClass *widget_class;

    widget_class = (GtkWidgetClass *)klass;
    widget_class->realize = sgwmtop_realize;
    widget_class->show = sgwmtop_show;
    widget_class->size_allocate = sgwmtop_size_allocate;
    widget_class->size_request = sgwmtop_size_request;
    widget_class->expose_event = sgwmtop_expose;
    GTK_CONTAINER_CLASS(klass)->check_resize = sgwmtop_check_resize;
}

static void sgwmtop_init(SGWMTop *sgwm)
{
    int i;

    sgwm->bar = sgwm->area = NULL;
    for (i = 0;i < SG_N_WORKSPACE;i++) {
	sgwm->workspace[i] = NULL;
	sgwm->curTask[i] = NULL;
	sgwm->frame[i] = NULL;
    }
    sgwm->unmappedTasks = NULL;
    sgwm->activeWorkSpace = -1;
    sgwm->position = -1;
    GTK_WIDGET_SET_FLAGS(GTK_WIDGET(sgwm),GTK_TOPLEVEL);
    GTK_WIDGET_UNSET_FLAGS(GTK_WIDGET(sgwm),GTK_NO_WINDOW);
    gtk_container_set_resize_mode(GTK_CONTAINER(sgwm),GTK_RESIZE_QUEUE);
}

GtkWidget *sgwmtop_new(SGWMWidget *sgwm)
{
    SGWMTop *top;

    top = SGWMTOP(gtk_type_new(sgwmtop_get_type()));
    top->wmwidget = sgwm;
    return GTK_WIDGET(top);
}

static void sgwmtop_do_map_task(SGWMTop *top, SGWindow *win, gint no)
{
    GtkWidget *winwidget;

    winwidget = GTK_WIDGET(win);
    win->workspace = no;
    if (SGWINDOW_DESTROYED(win)) return;
    if (no < 0 || no >= SG_N_WORKSPACE) {
	no = SG_N_WORKSPACE-1;
    }
    sgwindow_change_wm_state(win,NormalState);
    sgwindow_set_workspace_property(win,no);
    sgworkspace_add(GTK_CONTAINER(top->workspace[no]),winwidget);
}

static void sgwmtop_do_unmap_task(SGWMTop *top, SGWindow *win)
{
    GtkWidget *winwidget;

    winwidget = GTK_WIDGET(win);
    if (SGWINDOW_DESTROYED(win)) return;
    if (win->workspace < 0) return;
    if (win->workspace == top->activeWorkSpace) {
	sgbarholder_remove(GTK_CONTAINER(top->bar),winwidget);
    }
    if (win->workspace >= SG_N_WORKSPACE) {
	win->workspace = SG_N_WORKSPACE-1;
    }
    sgworkspace_remove(GTK_CONTAINER(top->workspace[win->workspace]),winwidget);
    sgwindow_change_wm_state(win,IconicState);
}

void sgwmtop_map_task_iconic(SGWMTop *top, SGWindow *win)
{
    GtkWidget *winwidget;

    winwidget = GTK_WIDGET(win);
    gtk_widget_ref(winwidget);
    win->workspace = 0;
    top->unmappedTasks = g_list_append(top->unmappedTasks,
      (gpointer)win);
}

void sgwmtop_map_nunome(SGWMTop *top, SGWindow *win)
{
    GtkWidget *winwidget;
    SGWorkSpace *wsp;
    gint no;

    winwidget = GTK_WIDGET(win);
    no = top->activeWorkSpace;
    no++;
    if (no >= SG_N_WORKSPACE) no = 0;
    if (win->workspace < 0) {
	gtk_widget_ref(winwidget);
    } else {
	if (top->curTask[no] == win) return;
	if (top->curTask[win->workspace] == win) {
	    sgwmtop_unmap_task(top,win);
	}
    }
    if (top->curTask[no] != NULL) {
	sgwmtop_do_unmap_task(top,top->curTask[no]);
	top->unmappedTasks = g_list_prepend(top->unmappedTasks,
	  (gpointer)top->curTask[no]);
    }
    sgwmtop_do_map_task(top,win,no);
    wsp = SGWORKSPACE(top->workspace[no]);
    wsp->nunome = TRUE;
    top->curTask[no] = win;
    if (!wsp->disabled) {
#ifdef __SIKIGAMI_PDA__
	top->position = top->frame[0]->allocation.height;
#else
	top->position = top->frame[0]->allocation.width;
#endif
    }
    nunome_set_position(top, win, no, 0);
    paned_chk_handle(top);
}

void sgwmtop_map_task(SGWMTop *top, SGWindow *win, gint no)
{
    GtkWidget *winwidget;
    SGWorkSpace *wsp;

    winwidget = GTK_WIDGET(win);
    if (no < 0) no = top->activeWorkSpace;
    if (no >= SG_N_WORKSPACE) {
	return;
    }
    if (no < 0) no = 0;
    wsp = SGWORKSPACE(top->workspace[no]);
    if (wsp->disabled || wsp->nunome) {
	no++;
	if (no >= SG_N_WORKSPACE) no = 0;
    }

    if (win->workspace < 0) {
	gtk_widget_ref(winwidget);
    } else {
	if (top->curTask[no] == win) return;
	if (top->curTask[win->workspace] == win) {
	    sgwmtop_unmap_task(top,win);
	}
    }
    if (top->curTask[no] != NULL) {
	sgworkspace_cover(top->workspace[no]);
	sgwmtop_do_unmap_task(top,top->curTask[no]);
	top->unmappedTasks = g_list_prepend(top->unmappedTasks,
	  (gpointer)top->curTask[no]);
    }
    if (win->workspace >= 0) {
	top->unmappedTasks = g_list_remove(top->unmappedTasks,
	  (gpointer)win);
    }
    sgwmtop_do_map_task(top,win,no);
    top->curTask[no] = win;
    sgwmtop_do_activate_workspace(top,no,TRUE,TRUE);
    sgworkspace_uncover(top->workspace[no]);
}

static SGWindow *sgwmtop_get_next_task(SGWMTop *top)
{
    SGWindow *nwin;

    nwin = NULL;
    while (top->unmappedTasks != NULL) {
	nwin = (SGWindow *)(top->unmappedTasks->data);
	top->unmappedTasks = g_list_remove(top->unmappedTasks,
	  top->unmappedTasks->data);
	sgwindow_ping(nwin);
	if (!SGWINDOW_DESTROYED(nwin)) break;
	gtk_widget_unref(GTK_WIDGET(nwin));
	nwin = NULL;
    }
    return nwin;
}

void sgwmtop_unmap_nunome(SGWMTop *top)
{
    int i;

    for (i = 0;i < SG_N_WORKSPACE;i++) {
	SGWorkSpace *wsp = SGWORKSPACE(top->workspace[i]);

	if (wsp->nunome) {
	    sgwmtop_unmap_task(top,top->curTask[i]);
	    break;
	}
    }
    paned_chk_handle(top);
    return;
}

void sgwmtop_popup_nunome()
{
    Window owner;
    XClientMessageEvent ev;

    gdk_error_trap_push();
    owner = XGetSelectionOwner(GDK_DISPLAY(),xa_nunome);
    if (owner != None) {
	ev.type = ClientMessage;
	ev.message_type = xa_nunome_popup;
	ev.format = 32;
	ev.window = owner;
	XSendEvent(GDK_DISPLAY(),owner,False,0,(XEvent *)&ev);
    }
    if (gdk_error_trap_pop()) {
	;
    }
}

void sgwmtop_switch_nunome(SGWMTop *top)
{
    int i;

    for (i = 0;i < SG_N_WORKSPACE;i++) {
	SGWorkSpace *wsp = SGWORKSPACE(top->workspace[i]);

	if (wsp->nunome) {
	    sgwmtop_unmap_task(top,top->curTask[i]);
	    return;
	}
    }
    sgwmtop_popup_nunome();
    paned_chk_handle(top);
    return;
}

void sgwmtop_unmap_task(SGWMTop *top, SGWindow *win)
{
    GtkWidget *winwidget;
    SGWindow *nwin;
    int no;
    SGWorkSpace *wsp;

    winwidget = GTK_WIDGET(win);
    if (win->workspace < 0) return;
    if (win->workspace >= SG_N_WORKSPACE) {
	win->workspace = SG_N_WORKSPACE-1;
    }
    no = win->workspace;
    if (top->curTask[no] != win) return;
    wsp = SGWORKSPACE(top->workspace[no]);
    XGrabServer(GDK_DISPLAY());
    if (wsp->disabled) {
	nwin = NULL;
    } else {
	nwin = sgwmtop_get_next_task(top);
    }
    sgworkspace_cover(top->workspace[no]);
    sgwmtop_do_unmap_task(top,win);
    if (wsp->nunome) {
	wsp->nunome = FALSE;
	if (wsp->disabled) {
	    gint np;

	    if (no == 0) {
		np = 0;
	    } else {
#ifdef __SIKIGAMI_PDA__
		np = GTK_WIDGET(top)->allocation.height;
#else
		np = GTK_WIDGET(top)->allocation.width;
#endif
	    }
	    paned_chk_handle(top);
	    sgwm_paned_set_position(GTK_PANED(top->area),np);
	} else {
	    paned_chk_handle(top);
	    sgwm_paned_set_position(GTK_PANED(top->area),top->position);
	}
    } else {
	top->unmappedTasks = g_list_prepend(top->unmappedTasks,(gpointer)win);
    }
    top->curTask[no] = nwin;
    if (nwin != NULL) {
	sgwmtop_do_map_task(top,nwin,no);
	if (top->activeWorkSpace == no) {
	    sgwmtop_do_activate_workspace(top,no,FALSE,TRUE);
	}
    } else if (top->activeWorkSpace == no) {
	int i;

	i = top->activeWorkSpace-1;
	if (i < 0) i = SG_N_WORKSPACE-1;
	while (i != no && top->curTask[i] == NULL) {
	    i--;
	    if (i < 0) i = SG_N_WORKSPACE-1;
	}
	if (i != no) {
	    sgwmtop_do_activate_workspace(top,i,TRUE,TRUE);
	} else {
	    sgwmtop_do_activate_workspace(top,0,TRUE,TRUE);
	}
    }
    XUngrabServer(GDK_DISPLAY());
    sgworkspace_uncover(top->workspace[no]);
}

int sgwmtop_get_workspace_coord(SGWMTop *top, int x, int y)
{
    int i;

    for (i = 0;i < SG_N_WORKSPACE;i++) {
	if (top->workspace[i] != 0
	  && sgworkspace_in(SGWORKSPACE(top->workspace[i]),x,y)) {
	    return i;
	}
    }
    return -1;
}

void sgwmtop_destroy_child_notify(SGWMTop *top, SGWindow *win)
{
    GtkWidget *winwidget = GTK_WIDGET(win);

    if (win->workspace < 0) return;
    if (g_list_find(top->unmappedTasks,(gpointer)win) != NULL) {
	top->unmappedTasks = g_list_remove(top->unmappedTasks,
	  (gpointer)win);
	gtk_widget_unref(winwidget);
    }
}

int sgwmtop_get_workspace_transient(SGWMTop *top, SGWindow *win)
{
    int i;

    for (i = 0;i < SG_N_WORKSPACE;i++) {
	if (top->curTask[i] != NULL) {
	    if (sgwindow_is_same_leader(top->curTask[i],win)) {
		int j = i;
		SGWorkSpace *ws;

		do {
		    j++;
		    if (j >= SG_N_WORKSPACE) j = 0;
		} while (j != i && ((ws = SGWORKSPACE(top->workspace[j]))->disabled
		  || ws->nunome));
		return j;
	    }
	}
    }
    return -1;
}

void sgwmtop_swap_task(SGWMTop *top, gint ws1, gint ws2)
{
    SGWindow *ct1,*ct2;
    int i;
    int aws;
    SGWorkSpace *wsp;

    if (ws1 < 0) ws1 = top->activeWorkSpace;
    if (ws2 < 0) ws2 = top->activeWorkSpace;
    if (ws1 < 0 || ws2 < 0) return;
    if (ws1 >= SG_N_WORKSPACE || ws2 >= SG_N_WORKSPACE) return;
    wsp = SGWORKSPACE(top->workspace[ws1]);
    if (wsp->disabled || wsp->nunome) return;
    if (ws1 == ws2) {
	do {
	    ws2++;
	    if (ws2 >= SG_N_WORKSPACE) ws2 = SG_N_WORKSPACE-1;
	    if (ws1 == ws2) return;
	    wsp = SGWORKSPACE(top->workspace[ws2]);
	} while (wsp->disabled || wsp->nunome);
    } else {
	wsp = SGWORKSPACE(top->workspace[ws2]);
	if (wsp->disabled || wsp->nunome) return;
    }
    ct1 = top->curTask[ws1];
    ct2 = top->curTask[ws2];
    aws = top->activeWorkSpace;
    sgworkspace_cover(top->workspace[ws1]);
    sgworkspace_cover(top->workspace[ws2]);
    if (ct1 != NULL) {
	gtk_widget_ref(GTK_WIDGET(ct1));
	sgwmtop_do_unmap_task(top,ct1);
    }
    if (ct2 != NULL) {
	gtk_widget_ref(GTK_WIDGET(ct2));
	sgwmtop_do_unmap_task(top,ct2);
    }
    if (ct1 != NULL) {
	sgwmtop_do_map_task(top,ct1,ws2);
	gtk_widget_unref(GTK_WIDGET(ct1));
    }
    top->curTask[ws2] = ct1;
    if (ct2 != NULL) {
	sgwmtop_do_map_task(top,ct2,ws1);
	gtk_widget_unref(GTK_WIDGET(ct2));
    }
    top->curTask[ws1] = ct2;
    i = aws;
    while (top->curTask[i] == NULL) {
	i--;
	if (i < 0) {
	    i = SG_N_WORKSPACE-1;
	}
	if (i == aws) break;
    }
    if (top->curTask[i] != NULL) {
	sgwmtop_do_activate_workspace(top,i,i != aws,TRUE);
    } else {
	sgwmtop_do_activate_workspace(top,0,aws != 0,TRUE);
    }
    sgworkspace_uncover(top->workspace[ws1]);
    sgworkspace_uncover(top->workspace[ws2]);
}

void sgwmtop_roll_up_task(SGWMTop *top, gint ws)
{
    SGWindow *win;
    SGWorkSpace *wsp;

    if (ws < 0) ws = top->activeWorkSpace;
    if (ws < 0) return;
    if (ws >= SG_N_WORKSPACE) return;
    wsp = SGWORKSPACE(top->workspace[ws]);
    if (wsp->disabled || wsp->nunome) return;
    if (top->unmappedTasks == NULL) return;
    win = top->curTask[ws];
    if (win != NULL) {
	sgwmtop_unmap_task(top,win);
	if (top->unmappedTasks != NULL
	  && (SGWindow *)top->unmappedTasks->data == win) {
	    top->unmappedTasks = g_list_remove(top->unmappedTasks,win);
	    top->unmappedTasks = g_list_append(top->unmappedTasks,win);
	}
    } else {
	sgwmtop_map_task(top,(SGWindow *)(top->unmappedTasks->data),ws);
    }
}

void sgwmtop_roll_down_task(SGWMTop *top, gint ws)
{
    SGWindow *win;
    SGWorkSpace *wsp;

    if (ws < 0) ws = top->activeWorkSpace;
    if (ws < 0) return;
    if (ws >= SG_N_WORKSPACE) return;
    wsp = SGWORKSPACE(top->workspace[ws]);
    if (wsp->disabled || wsp->nunome) return;
    if (top->unmappedTasks == NULL) return;
    win = (SGWindow *)(g_list_last(top->unmappedTasks)->data);
    sgwmtop_map_task(top, win, ws);
}

void sgwmtop_move_next_task(SGWMTop *top, gint ws)
{
    SGWindow *win;
    SGWorkSpace *wsp;

    if (ws < 0) ws = top->activeWorkSpace;
    if (ws < 0) return;
    if (ws >= SG_N_WORKSPACE) return;
    wsp = SGWORKSPACE(top->workspace[ws]);
    if (wsp->disabled || wsp->nunome) return;
    if ((win = top->curTask[ws]) == NULL) return;
    sgwmtop_unmap_task(top,win);
    ws++;
    if (ws >= SG_N_WORKSPACE) ws = 0;
    sgwmtop_map_task(top,win,ws);
}

void sgwmtop_move_prev_task(SGWMTop *top, gint ws)
{
    SGWindow *win;
    SGWorkSpace *wsp;

    if (ws < 0) ws = top->activeWorkSpace;
    if (ws < 0) return;
    if (ws >= SG_N_WORKSPACE) return;
    wsp = SGWORKSPACE(top->workspace[ws]);
    if (wsp->disabled || wsp->nunome) return;
    if ((win = top->curTask[ws]) == NULL) return;
    sgwmtop_unmap_task(top,win);
    ws--;
    if (ws < 0) ws = SG_N_WORKSPACE-1;
    sgwmtop_map_task(top,win,ws);
}

void sgwmtop_enable_workspace(SGWMTop *top, gint ws, gint enable)
{
    SGWorkSpace *wsp;
    SGWindow *nwin;

    if (ws < 0) ws = top->activeWorkSpace;
    if (ws < 0 || ws >= SG_N_WORKSPACE) return;
    wsp = SGWORKSPACE(top->workspace[ws]);
    if (enable < 0) {
	enable = wsp->disabled ? 1 : 0;
    }
    if (enable == 0) {
	if (wsp->disabled || wsp->nunome) return;
	if(sgwmtop_get_enabled_workspace_count(top, NULL) <= 1) return;
	wsp->disabled = TRUE;
#ifdef __SIKIGAMI_PDA__
	top->position = top->frame[0]->allocation.height;
#else
	top->position = top->frame[0]->allocation.width;
#endif
	if (top->curTask[ws] != NULL) {
	    sgwmtop_unmap_task(top,top->curTask[ws]);
	}
	workspace_disable(top, ws);
	paned_chk_handle(top);
    } else {
	if (!wsp->disabled || wsp->nunome) return;
	wsp->disabled = FALSE;
	sgwm_paned_set_position(GTK_PANED(top->area),top->position);
	paned_chk_handle(top);
	if ((nwin = sgwmtop_get_next_task(top)) != NULL) {
	    sgwmtop_map_task(top,nwin,ws);
	}
    }
}

gint sgwmtop_get_enabled_workspace_count(SGWMTop *top, gint *lastOne)
{
    int i;
    gint n;
    SGWorkSpace *wsp;

    if(lastOne != NULL) *lastOne = -1;
    n = 0;
    for (i = 0;i < SG_N_WORKSPACE;i++) {
	wsp = SGWORKSPACE(top->workspace[i]);
	if (!wsp->disabled && !wsp->nunome) {
	    n++;
	    if(lastOne != NULL) *lastOne = i;
	}
    }
    return n;
}

void sgwmtop_resize_workspace(SGWMTop *top, gint ws, gint mode, gint size)
{
    gint lastOne;
    GtkWidget *widget;
    SGWorkSpace *wsp;

    if (ws < 0) ws = top->activeWorkSpace;
    if (ws >= SG_N_WORKSPACE) return;
    if (ws < 0) ws = 0;
    wsp = SGWORKSPACE(top->workspace[ws]);
    if (wsp->disabled || wsp->nunome) return;
    if (sgwmtop_get_enabled_workspace_count(top,&lastOne) <= 1) {
	return;
    }
    if (mode != 0) {
#ifdef __SIKIGAMI_PDA__
	size += GTK_WIDGET(top->frame[ws])->allocation.height;
#else
	size += GTK_WIDGET(top->frame[ws])->allocation.width;
#endif
    }
    if (size < 0) size = 0;
    widget = GTK_WIDGET(top);
#ifdef __SIKIGAMI_PDA__
    if (size > widget->allocation.width) size = widget->allocation.height;
    if (ws != 0) {
	size = widget->allocation.height-size;
    }
#else
    if (size > widget->allocation.width) size = widget->allocation.width;
    if (ws != 0) {
	size = widget->allocation.width-size;
    }
#endif
    sgwm_paned_set_position(GTK_PANED(top->area),size);
}

void sgwmtop_calculate_win_size(SGWMTop *top, SGWindow *sgwin,
  GtkAllocation *newalloc)
{
    if (sgwin->workspace < 0) return;
    if (top->curTask[sgwin->workspace] != sgwin) return;
    sgworkspace_calculate_child_size(SGWORKSPACE(top->workspace[sgwin->workspace]),
      sgwin,newalloc);
}

void sgwmtop_call_desktop(SGWMTop *top, SGWindow *sgwin, gint ws)
{
    if (ws < 0) {
	ws = top->activeWorkSpace;
	if (ws < 0) ws = 0;
    }
    if (ws >= SG_N_WORKSPACE) {
	return;
    }
    if (top->curTask[ws] == sgwin) {
	sgwmtop_unmap_task(top,sgwin);
    } else {
	sgwmtop_map_task(top,sgwin,ws);
    }
}
